# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
from typing import Any
from typing import Awaitable

import fastapi

from cbra.types import IDependant
from cbra.types import IVerifier
from cbra.types import Request


DEFAULT_API_VERSION: str = '2023-01'


class ShopifyWebhookMessage(IDependant):
    __module__: str = 'cbra.ext.shopify'
    api_version: str
    content: dict[str, Any]
    domain: str | None
    hmac_sha256: bytes | None = None
    topic: str | None = None
    webhook_id: str | None = None

    def __init__(
        self,
        api_version: str | None = fastapi.Header(
            default=None,
            alias='X-Shopify-API-Version',
        ),
        domain: str | None = fastapi.Header(
            default=None,
            alias='X-Shopify-Shop-Domain',
        ),
        signature: str | None = fastapi.Header(
            default=None,
            alias='X-Shopify-Hmac-Sha256',
        ),
        topic: str | None = fastapi.Header(
            default=None,
            alias='X-Shopify-Topic'
        ),
        webhook_id: str | None = fastapi.Header(
            default=None,
            alias='X-Shopify-Webhook-Id'
        ),
        content: dict[str, Any] = fastapi.Body()
    ):
        self.api_version = api_version or DEFAULT_API_VERSION
        self.content = content
        self.domain = domain
        self.topic = topic
        self.webhook_id = webhook_id
        if signature is not None:
            self.hmac = base64.b64decode(str.encode(signature, 'utf-8'))

    def verify(self, request: Request, verifier: IVerifier) -> bool | Awaitable[bool]:
        return verifier.verify(self.hmac, request.sha256)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

from cbra.types import NullRequestPrincipal
from cbra.types import RequestPrincipal
from cbra.types import OIDCRequestPrincipal
from cbra.types import RFC9068RequestPrincipal


def test_parse_bearer_oidc():
    token = (
        'eyJ0eXAiOiJKV1QiLCJhbGciOiJFUzI1NiIsImtpZCI6IjgwNDFiM2VmYWRlYm'
        'VlMjNhOWY5MTZhZjU0MDBhYmJlIn0.eyJhdWQiOiJodHRwczovL2xvY2FsaG9z'
        'dDo4MDAwIiwiYXpwIjoiY2xpZW50X2lkIiwiZW1haWwiOiJ1c2VyQGV4YW1wbG'
        'UuY29tIiwiZW1haWxfdmVyaWZpZWQiOnRydWUsImV4cCI6MTY3Njg5MTI5Miwi'
        'aWF0IjoxNjc2ODg3NjkyLCJpc3MiOiJodHRwczovL2NicmEiLCJzdWIiOiIxMj'
        'MifQ.R0KqeCRFNc4dRJCf6lBkH6AviYNa-t36xHkQD1vkG5eUaFJyB2OUjvQVS'
        'jP-QfHmoRUj5yV8zaBOpz1rl6fXHw'
    )
    principal = RequestPrincipal.parse_obj({
        'headers': {'Authorization': f'Bearer {token}'}
    })
    assert isinstance(principal.__root__, OIDCRequestPrincipal)


def test_parse_bearer_rfc9068():
    token = (
        'eyJ0eXAiOiJhdCtqd3QiLCJhbGciOiJFUzI1NiIsImtpZCI6IjgwNDFiM2VmYWRl'
        'YmVlMjNhOWY5MTZhZjU0MDBhYmJlIn0.eyJpc3MiOiJodHRwczovL2FjY291bnRz'
        'LndlYmlkZW50aXR5LmlkIiwic3ViIjoiY2xpZW50X2lkIiwiYXVkIjoiaHR0cHM6'
        'Ly9sb2NhbGhvc3Q6ODAwMCIsImV4cCI6MTY3Njg4ODIzMSwibmJmIjoxNjc2ODg3'
        'OTMxLCJpYXQiOjE2NzY4ODc5MzEsImp0aSI6IkJleGE1RzlZX05RZW83SXd0MU4z'
        'N0l1WENiWUkxQlo1IiwiY2xpZW50X2lkIjoiY2xpZW50X2lkIn0.XI18HDgpu8YI'
        'GHASGuTkOb2hnI9lDBZHNlFl-Tsl7fXeHbyc7IMvw1o7h4MeEFEubVPeTq-FHZzo'
        'aX0Wea6QWQ'
    )
    principal = RequestPrincipal.parse_obj({
        'headers': {'Authorization': f'Bearer {token}'}
    })
    assert isinstance(principal.__root__, RFC9068RequestPrincipal)



@pytest.mark.parametrize("value", [
    'Bearer '
    'aeyJ0eXAiOiJhdCtqd3QiLCJhbGciOiJFUzI1NiIsImtpZCI6IjgwNDFiM2VmYWRl'
    'YmVlMjNhOWY5MTZhZjU0MDBhYmJlIn0.eyJpc3MiOiJodHRwczovL2FjY291bnRz'
    'LndlYmlkZW50aXR5LmlkIiwic3ViIjoiY2xpZW50X2lkIiwiYXVkIjoiaHR0cHM6'
    'Ly9sb2NhbGhvc3Q6ODAwMCIsImV4cCI6MTY3Njg4ODIzMSwibmJmIjoxNjc2ODg3'
    'OTMxLCJpYXQiOjE2NzY4ODc5MzEsImp0aSI6IkJleGE1RzlZX05RZW83SXd0MU4z'
    'N0l1WENiWUkxQlo1IiwiY2xpZW50X2lkIjoiY2xpZW50X2lkIn0.XI18HDgpu8YI'
    'GHASGuTkOb2hnI9lDBZHNlFl-Tsl7fXeHbyc7IMvw1o7h4MeEFEubVPeTq-FHZzo'
    'aX0Wea6QWQ',
    'Bearer foo',
    'Foo',
])
def test_parse_bearer_malformed_becomes_nullrequestprincipal(value: str):
    token = (
    )
    principal = RequestPrincipal.parse_obj({
        'headers': {'Authorization': token}
    })
    assert isinstance(principal.__root__, NullRequestPrincipal)


def test_bearer_untyped_jwt_is_nullrequestprincipal():
    token = (
        'eyJhbGciOiJFUzI1NiIsImtpZCI6IjgwNDFiM2VmYWRlYmVlMjNhOWY5MTZhZjU0'
        'MDBhYmJlIn0.eyJpc3MiOiJodHRwczovL2FjY291bnRzLndlYmlkZW50aXR5Lmlk'
        'Iiwic3ViIjoiY2xpZW50X2lkIiwiYXVkIjoiaHR0cHM6Ly9sb2NhbGhvc3Q6ODAw'
        'MCIsImV4cCI6MTY3Njg4ODIzMSwibmJmIjoxNjc2ODg3OTMxLCJpYXQiOjE2NzY4'
        'ODc5MzEsImp0aSI6IkJleGE1RzlZX05RZW83SXd0MU4zN0l1WENiWUkxQlo1Iiwi'
        'Y2xpZW50X2lkIjoiY2xpZW50X2lkIn0.P54GYkelEWnQGWgg8qP_CMZaOI_rlWEP'
        'wU1dh5WkZRL5Dx4jly94_wO6HTW4If1VzPKWFJiDNPNUqDr650iq_Q'
    )
    principal = RequestPrincipal.parse_obj({
        'headers': {'Authorization': f'Bearer {token}'}
    })
    assert isinstance(principal.__root__, NullRequestPrincipal)


def test_bearer_unaccepted_jwt_is_nullrequestprincipal():
    token = (
        'eyJhbGciOiJFUzI1NiIsInR5cCI6ImZvbyIsImtpZCI6IjgwNDFiM2VmYWRlYmVl'
        'MjNhOWY5MTZhZjU0MDBhYmJlIn0.eyJpc3MiOiJodHRwczovL2FjY291bnRzLndl'
        'YmlkZW50aXR5LmlkIiwic3ViIjoiY2xpZW50X2lkIiwiYXVkIjoiaHR0cHM6Ly9s'
        'b2NhbGhvc3Q6ODAwMCIsImV4cCI6MTY3Njg4ODIzMSwibmJmIjoxNjc2ODg3OTMx'
        'LCJpYXQiOjE2NzY4ODc5MzEsImp0aSI6IkJleGE1RzlZX05RZW83SXd0MU4zN0l1'
        'WENiWUkxQlo1IiwiY2xpZW50X2lkIjoiY2xpZW50X2lkIn0.U9iSF3Kluw1gBFAL'
        'IdEJVwwGCHO_DS5JZlqs_8ynA8srOjY00hYB2cWLeJ6RIwcNRE2IPsSkwbmP6kGD'
        '6ym0wA'
    )
    principal = RequestPrincipal.parse_obj({
        'headers': {'Authorization': f'Bearer {token}'}
    })
    assert isinstance(principal.__root__, NullRequestPrincipal)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from datetime import datetime
from datetime import timezone
from typing import Any
from typing import TypeVar

import pydantic
from cbra.types import PersistedModel
from ..types import QueryAuthorizeResponse


T = TypeVar('T', bound='AuthorizationState')


class AuthorizationState(PersistedModel):
    created: datetime = pydantic.Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )
    nonce: str
    redirect_uri: str
    state: str = pydantic.Field(
        default=...,
        primary_key=True
    )

    @classmethod
    def new(cls: type[T], redirect_uri: str) -> T:
        return cls(
            nonce=secrets.token_urlsafe(48),
            redirect_uri=redirect_uri,
            state=secrets.token_urlsafe(48),
        )
    
    def is_valid(self, params: QueryAuthorizeResponse | Any) -> bool:
        return isinstance(params, QueryAuthorizeResponse) and params.state == self.state


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any
from typing import AsyncGenerator

from canonical import EmailAddress
from google.cloud.datastore import Query
from headless.ext.oauth2.models import SubjectIdentifier

from cbra.types import IModelRepository
from cbra.core.iam import ISubjectRepository
from cbra.core.iam.models import Principal
from cbra.core.iam.models import Subject
from .datastorerepository import DatastoreRepository


class DatastoreSubjectRepository(
    DatastoreRepository,
    ISubjectRepository,
    IModelRepository[Subject]
):
    __module__: str = 'cbra.ext.google'

    @functools.singledispatchmethod # type: ignore
    async def get(self, principal: Any) -> Subject | None:
        if principal is None: return None
        raise NotImplementedError(f'{principal.__module__}.{type(principal).__name__}')

    @get.register
    async def get_by_email(self, principal: EmailAddress) -> Subject | None:
        query = self.query('Principal')
        query.add_filter('spec.kind', '=', 'EmailAddress') # type: ignore
        query.add_filter('spec.email', '=', str(principal)) # type: ignore
        dao = await self.one(Principal, query)
        if dao is None:
            return None
        return await self.get(dao.subject)

    @get.register
    async def get_by_identifier(self, principal: SubjectIdentifier) -> Subject | None:
        query = self.query('Principal')
        query.add_filter('spec.kind', '=', 'SubjectIdentifier') # type: ignore
        query.add_filter('spec.iss', '=', principal.iss) # type: ignore
        query.add_filter('spec.sub', '=', principal.sub) # type: ignore
        dao = await self.one(Principal, query)
        if dao is None:
            return None
        return await self.get(dao.subject)

    @get.register
    async def get_by_uid(self, uid: int) -> Subject | None:
        assert isinstance(uid, int), repr(uid)
        obj = await self.get_model_by_key(Subject, uid)
        if obj is None:
            return None
        assert obj.uid is not None
        async for principal in self.get_principals(obj.uid):
            obj.principals.add(principal)
        return obj

    async def get_principals(self, subject_id: int) -> AsyncGenerator[Principal, None]:
        query = self.query(Principal)
        query.add_filter('subject', '=', subject_id) # type: ignore
        for entity in await self.run_in_executor(query.fetch): # type: ignore
            yield self.restore(Principal, entity)

    async def resolve(self, identity: int | str) -> Subject | None:
        """Resolve a principal to a global subject identifier."""
        return await self.get(int(identity))

    async def find_by_principals(
        self,
        principals: list[Any]
    ) -> set[int]:
        subjects: set[int] = set()
        for principal in principals:
            query = self.query(Principal)
            self.filter_principal(principal, query)
            subjects.update([
                int(x['subject']) async for x in self.execute(query, dict[str, int])
            ])
        return subjects

    async def persist(self, instance: Subject) -> Subject:
        await self.put(instance, exclude={'principals', 'uid'})
        assert instance.uid is not None
        await self.delete([self.model_key(x) for x in instance._removed_principals]) # type: ignore
        for principal in instance.principals:
            await self.put(principal, exclude={'id'})
        return instance

    @functools.singledispatchmethod
    def filter_principal(
        self,
        principal: EmailAddress | SubjectIdentifier,
        query: Query
    ) -> None:
        raise NotImplementedError(type(principal).__name__)

    @filter_principal.register
    def filter_emailaddress(
        self,
        principal: EmailAddress,
        query: Query
    ) -> None:
        query.add_filter('spec.kind', '=', 'EmailAddress') # type: ignore
        query.add_filter('spec.email', '=', str(principal)) # type: ignore

    @filter_principal.register
    def filter_subjectidentifier(
        self,
        principal: SubjectIdentifier,
        query: Query
    ) -> None:
        query.add_filter('spec.kind', '=', 'SubjectIdentifier') # type: ignore
        query.add_filter('spec.iss', '=', principal.iss) # type: ignore
        query.add_filter('spec.sub', '=', principal.sub) # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hmac

from cbra.types import IVerifier


class HMACWebhookVerifier(IVerifier):
    __module__: str = 'cbra.ext.webhooks'
    algorithm: str

    def __init__(self, secret: bytes, algorithm: str):
        self.algorithm = algorithm
        self.__secret = secret

    async def verify(self, signature: bytes, message: bytes) -> bool:
        return hmac.compare_digest(
            hmac.digest(self.__secret, message, self.algorithm),
            signature
        )
    

class HMACSHA256WebhookVerifier(HMACWebhookVerifier):
    __module__: str = 'cbra.ext.webhooks'

    def __init__(self, secret: bytes):
        super().__init__(secret=secret, algorithm='sha256')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timezone

import pydantic
from canonical import EmailAddress

from cbra.types import PersistedModel


ALLOWED_DOMAINS: set[str] = {
    'gmail.com',
    'outlook.com',
    'hotmail.com',
    'live.com',
    'myyahoo.com',
    'yahoo.com',
    'rocketmail.com',
    'icloud.com',
    'pm.me',
    'protonmail.com'
}


class AccountRegistration(PersistedModel):
    """Retains the state of an account registration request that
    requires a backup email address.

    Registration of a new account is done by verifying an email address
    using a shared secret (verification code). If the email adres is not
    whitelist per its domain, then a fallback email address must be
    provided. The :class:`AccountRegistration` model must then resolve
    either of three cases:

    - Both email addresses are not linked to existing accounts. In this
      case, an account that owns both addresses is created.
    - One address is linked to an existing account. The other address is
      then also added to the account.
    - The email addresses are linked to different accounts. This is an
      error state.
    """
    id: int | None = pydantic.Field(
        default=None,
        auto_assign=True
    )

    requested: datetime = pydantic.Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )

    register: EmailAddress = pydantic.Field(
        default=...,
        primary_key=True
    )
    fallback: EmailAddress | None = None


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .basewebhookendpoint import BaseWebhookEndpoint
from .hmacwebhookverifier import HMACWebhookVerifier
from .types import IWebhookEnvelope
from .webhookendpoint import WebhookEndpoint
from .webhookendpointtype import WebhookEndpointType
from .webhookenvelope import WebhookEnvelope
from .webhookresponse import WebhookResponse


__all__: list[str] = [
    'BaseWebhookEndpoint',
    'HMACWebhookVerifier',
    'IWebhookEnvelope',
    'WebhookEndpoint',
    'WebhookEndpointType',
    'WebhookEnvelope',
    'WebhookResponse'
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timezone
from typing import TypeVar

import pydantic


T = TypeVar('T', bound='ModelMetadata')


class ModelMetadata(pydantic.BaseModel):
    created: datetime = pydantic.Field(
        default=...
    )

    deleted: datetime | None = pydantic.Field(
        default=None
    )

    generation: int = pydantic.Field(
        default=0,
    )

    uid: int | str | None = pydantic.Field(
        default=None
    )

    updated: datetime = pydantic.Field(
        default=...
    )

    @classmethod
    def initialize(cls: type[T]) -> T:
        now = datetime.now(timezone.utc)
        return cls.parse_obj({
            'created': now,
            'updated': now,
        })


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from typing import Any

import fastapi

from cbra.types import RequestPrincipal
from cbra.types import IRoutable
from ..iam import AuthorizationContextFactory
from ..ioc import instance
from ..messagepublisher import MessagePublisher
from ..sessions import RequestSession
from .iresource import IResource
from .resourcemodel import ResourceModel
from .resourcetype import ResourceType


class Resource(IResource, metaclass=ResourceType):
    __abstract__: bool = True
    __actions__: list[type[IRoutable]] = []
    __module__: str = 'cbra.core'
    context_factory: AuthorizationContextFactory = AuthorizationContextFactory.depends()
    model: type[ResourceModel]
    principal: RequestPrincipal = RequestPrincipal.depends()
    publisher: MessagePublisher = instance('MessagePublisher')
    session: RequestSession = RequestSession.depends()
    timestamp: datetime

    def __init_subclass__(cls, model: type[ResourceModel]):
        cls.model = model

    @classmethod
    def add_to_router(cls, router: fastapi.FastAPI, **kwargs: Any) -> None:
        for action in cls.__actions__:
            kwargs.setdefault('response_model_by_alias', cls.response_model_by_alias)
            action.add_to_router(cls, router, **kwargs)

    def adapt(self, obj: Any) -> ResourceModel:
        """Adapt an object of any kind to a :class:`ResourceModel` instance.
        The default implementation simply returns the object.
        """
        return obj
    
    def parse_obj(self, obj: Any) -> Any:
        return self.model.parse_obj(obj)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import collections
from typing import TypeVar

from .basestorage import BaseStorage
from .models import AuthorizationState
from .models import AuthorizationServerModel


T = TypeVar('T')


class MemoryStorage(BaseStorage):
    """An in-memory implementation of :class:`~cbra.ext.oauth2.BaseStorage`."""
    __module__: str = 'cbra.ext.oauth2'
    objects: dict[str, dict[str, AuthorizationServerModel]] = collections.defaultdict(dict)

    @classmethod
    def clear(cls) -> None:
        cls.objects = collections.defaultdict(dict)

    def __init__(self):
        self.objects = MemoryStorage.objects

    async def destroy(self, obj: AuthorizationServerModel) -> None:
        self.objects[type(obj).__name__].pop(obj.__key__[0][1], None) # type: ignore

    async def get_state(self, key: str) -> AuthorizationState | None:
        return self._get(AuthorizationState, key)

    async def persist_state(self, obj: AuthorizationState) -> None:
        self.objects[type(obj).__name__][obj.state] = obj

    def _get(self, Model: type[T], key: str) -> T | None:
        return self.objects[Model.__name__].get(key)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Awaitable

import fastapi

from .iauthorizationcontext import IAuthorizationContext
from .irequestprincipal import IRequestPrincipal


class IAuthorizationContextFactory:
    """Setup an authorization context for a request."""
    __module__: str = 'cbra.types'

    async def authenticate(
        self,
        request: fastapi.Request,
        principal: IRequestPrincipal,
        providers: set[str] | None = None,
        subjects: set[str] | Awaitable[set[str]] | None  = None,
        claims: dict[str, Any] | None = None
    ) -> IAuthorizationContext:
        raise NotImplementedError

    def validate_audience(self, principal: IRequestPrincipal, allow: set[str]) -> None:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cbra.ext.webhooks import WebhookEndpoint

from .shopifywebhookmessage import ShopifyWebhookMessage


class ShopifyWebhookEndpoint(WebhookEndpoint):
    __module__: str = 'cbra.ext.shopify'
    domain: str = "shopify.com"
    envelope: ShopifyWebhookMessage
    require_authentication: bool = False
    summary: str = "Shopify"
    tags: list[str] = ["Webhooks (Incoming)"]

    async def on_order_created(
        self,
        envelope: ShopifyWebhookMessage
    ) -> None:
        print(self.request.sha256)
        print(
            envelope.api_version,
            envelope.domain,
            envelope.hmac_sha256,
            envelope.topic,
            envelope.webhook_id,
            envelope.content
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Awaitable

from cbra.types import IVerifier
from cbra.types import Request


class WebhookEnvelope:
    __module__: str = 'cbra.ext.webhooks'

    @property
    def event_name(self) -> str:
        return self.get_event_name()
    
    def get_event_name(self) -> str:
        """Subclasses must override this method to return the event name
        of the incoming message.
        """
        raise NotImplementedError("Subclasses must override this method.")
    

    def verify(
        self,
        request: Request,
        verifier: IVerifier
    ) -> Awaitable[bool]:
        """Subclasses must override this method to verify the authenticity "
        and integrity of the webhook message.
        """
        raise NotImplementedError("Subclasses must override this method.")


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic
from headless.ext import oauth2

from .responsevalidationfailure import ResponseValidationFailure


class QueryAuthorizeResponse(pydantic.BaseModel):
    code: str
    state: str | None = None

    async def obtain(
        self,
        client: oauth2.Client,
        state: str | None,
        **kwargs: Any
    ) -> oauth2.TokenResponse:
        if state != self.state:
            raise ResponseValidationFailure(
                "The state parameters do not match."
            )
        return await client.token(
            oauth2.AuthorizationEndpointResponse.parse_obj(self.dict()).__root__,
            **kwargs
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import AsyncGenerator

from cbra.types import IModelRepository
from .models import Subject
from .models import Principal


class ISubjectRepository(IModelRepository[Subject]):
    __module__: str = 'cbra.core.iam'
    model: type[Subject] = Subject

    def get_principals(self, subject_id: int) -> AsyncGenerator[Principal, None]:
        """Return the list of principals that are registered for a
        subject.
        """
        raise NotImplementedError

    async def find_by_principals(
        self,
        principals: list[Any]
    ) -> set[int]:
        """Identify subjects by the given principals and return a
        list of integers representing the subject identifiers.
        """
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from typing import Any
from typing import Literal

import pydantic

from cbra.types import ISessionManager
from cbra.types import PersistedModel
from ..types import PrincipalType
from .principal import Principal


class Subject(PersistedModel):
    kind: Literal['User']
    uid: int | None = pydantic.Field(
        default=None,
        auto_assign=True
    )

    created: datetime
    seen: datetime
    active: bool = True

    principals: set[Principal] = set()

    #: TODO: The PeristedModel must be refactored to register attribute mutations
    #: so that the storage backend can detect if it needs to delete something.
    _removed_principals: list[Principal] = pydantic.PrivateAttr([])

    def add_principal(
        self,
        issuer: str,
        value: PrincipalType,
        asserted: datetime,
        trust: bool = False
    ) -> None:
        assert self.uid is not None
        new = Principal.new(self.uid, issuer, value, asserted=asserted, trust=trust)
        old = None
        if new in self.principals:
            # TODO: ugly
            principals = list(self.principals)
            old = principals[principals.index(new)]
        must_add = any([
            old is None,
            old is not None and not old.trust
        ])
        if must_add:
            if old is not None:
                self._removed_principals.append(old)
                self.principals.remove(old)
            self.principals.add(new)

    def add_to_session(self, session: ISessionManager[Any]) -> None:
        raise NotImplementedError
    
    def has_principal(self, principal: PrincipalType) -> bool:
        p = Principal.new(
            0,
            '',
            principal,
            asserted=datetime.now(),
            trust=False
        )
        return p in self.principals


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
from typing import TypeVar

import pydantic

from .ihashable import IHashable
from .sessionclaims import SessionClaims


T = TypeVar('T', bound='SessionModel')


class SessionModel(pydantic.BaseModel, IHashable):
    id: str
    iat: int
    claims: SessionClaims | None = None
    hmac: str | None = None

    @classmethod
    def parse_cookie(cls: type[T], value: str | None) -> T | None:
        if value is None: return None
        try:
            serialized = base64.urlsafe_b64decode(str.encode(value, 'ascii'))
            return cls.parse_raw(serialized)
        except Exception:
            return None

    def digest(self) -> bytes:
        return self._hash(self.dict(exclude={'hmac'}, exclude_none=True))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
.. _ref-guides-integrating-an-oauth2-authorization-server

=============================================
Integrating an OAuth 2.x authorization server
=============================================
"""
import cbra.core as cbra
from cbra.core.conf import settings
from cbra.core.iam import ISubjectRepository
from cbra.core.iam.types import IUserOnboardingService
from cbra.core.iam.services import UserOnboardingService
from .memorystorage import MemoryStorage
from .types import IAuthorizationServerStorage


class AuthorizationServerEndpoint(cbra.Endpoint):
    onboard: IUserOnboardingService = cbra.ioc.instance(
        name='SubjectOnboardingService',
        missing=UserOnboardingService
    )
    storage: IAuthorizationServerStorage = cbra.ioc.instance(
        name='AuthorizationServerStorage',
        missing=MemoryStorage
    )
    subjects: ISubjectRepository = cbra.ioc.instance('SubjectRepository')

    def get_issuer(self) -> str:
        return settings.OAUTH2_ISSUER or\
            f'{self.request.url.scheme}://{self.request.url.netloc}'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import collections
import inspect

import pydantic.fields
from .basemodel import BaseModel
from .persistedmodel import PersistedModel


class ModelInspector:
    __module__: str = 'cbra.types'
    model: type[BaseModel]

    def get_column_fields(
        self,
        model: type[BaseModel] | BaseModel
    ) -> dict[str, pydantic.fields.ModelField]:
        """Return a mapping for field names for :class:`pydantic.field.ModelField`
        instances representing the fields that are not first-class storage
        entities i.e. having their own table or being a distinct kind.
        """
        return collections.OrderedDict([
            (k, v)
            for k, v in model.__fields__.items()
            if not self.is_first_class_entity(v.type_)
        ])
    
    def has_children(self, model: type[BaseModel] | BaseModel) -> bool:
        """Return a boolean indicating if the model has child entities."""
        return any([
            self.is_first_class_entity(field.type_)
            for field in model.__fields__.values()
        ])
    
    def auto_increments(self, obj: type[BaseModel] | BaseModel) -> bool:
        """Return a boolean indicating if the model has an auto incrementing
        field.
        """
        raise NotImplementedError
    
    def has_identity(self, obj: type[BaseModel] | BaseModel) -> bool:
        """Return a boolean indicating if the model has an identity."""
        raise NotImplementedError

    def is_first_class_entity(self, obj: type) -> bool:
        """Return a boolean indicating if the type is a first-class
        entity.
        """
        return inspect.isclass(obj) and issubclass(obj, BaseModel)
    
    def is_insert(self, obj: PersistedModel) -> bool:
        """Return a boolean indicating if the object must be inserted
        into the storage backend.
        """
        return obj.__metadata__.uid is None


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import Any

from .injected import Injected


class Instance(Injected):

    @property
    def __signature__(self) -> inspect.Signature:
        assert self.ref is not None, self.name
        return inspect.signature(self.factory)

    def callable(self) -> bool:
        return True

    async def resolve(self, *args: Any, **kwargs: Any) -> Any:
        assert self.ref is not None
        return self.ref.init(*args, **kwargs)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from .loader import import_symbol


class Dependency(pydantic.BaseModel):
    name: str
    qualname: str
    symbol: Any = None
    instance: Any = None
    args: list[Any] = []
    kwargs: dict[str, Any] = {}
    singleton: bool = False

    def get(self):
        return self.symbol

    def init(self, *args: Any, **kwargs: Any) -> Any:
        assert self.symbol is not None
        return self.instance or self.symbol(*args, **kwargs)

    def resolve(self) -> None:
        try:
            self.symbol = import_symbol(self.qualname)
        except ImportError:
            raise
        if self.singleton:
            self.instance = self.symbol(*self.args, **self.kwargs)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any

from google.cloud.datastore import Client
from google.cloud.datastore import Entity

from .runner import Runner
from .types import IDatastoreCursor
from .types import IDatastoreEntity
from .types import IDatastoreKey
from .types import IDatastoreQuery


class BaseDatastoreRepository(Runner):
    __module__: str = 'cbra.ext.google'

    def __init__(
        self,
        client: Client
    ) -> None:
        self.client = client

    def key(
        self,
        kind: str,
        identifier: int | str | None = None,
        parent: IDatastoreKey | None = None
    ) -> IDatastoreKey:
        args: list[Any] = [kind]
        if identifier is not None:
            args.append(identifier)
        return self.client.key(*args, parent=parent) # type: ignore
    
    def entity(
        self,
        kind: str,
        key: int | str | IDatastoreKey | None = None,
        parent: IDatastoreKey | None = None
    ) -> IDatastoreEntity:
        if isinstance(key, (int, str)):
            key = self.key(kind, key)
        return self.client.entity(key or self.key(kind, parent=parent)) # type: ignore

    def query(self, *args: Any, **kwargs: Any) -> IDatastoreQuery:
        return self.client.query(*args, **kwargs) # type: ignore

    async def execute(self, query: IDatastoreQuery) -> IDatastoreCursor:
        cursor = await self.run_in_executor(functools.partial(query.fetch))
        return cursor

    async def get_entity_by_key(self, key: IDatastoreKey) -> Entity | None:
        return await self.run_in_executor(
            functools.partial(
                self.client.get, # type: ignore
                key=key
            )
        )

    async def put(self, entity: IDatastoreEntity) -> IDatastoreKey:
        await self.run_in_executor(self.client.put, entity) # type: ignore
        return entity.key # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets

import pytest
from google.cloud.datastore import Client

from cbra.ext.google import BaseDatastoreRepository


@pytest.mark.asyncio
async def test_put_autogenerated_and_retrieve(
    client: Client
):
    repo = BaseDatastoreRepository(client=client)
    e1 = repo.entity('Test')
    e1['foo'] = 'foo'
    k1 = await repo.put(e1)
    assert k1 == e1.key
    e2 = await repo.get_entity_by_key(e1.key)
    assert e1 == e2


@pytest.mark.asyncio
async def test_put_autogenerated_and_query(
    client: Client
):
    repo = BaseDatastoreRepository(client=client)
    e1 = repo.entity('Test')
    e1['foo'] = 'foo'
    k1 = await repo.put(e1)
    assert k1.id is not None
    q = repo.query(kind='Test')
    q.add_filter('foo', '=', 'foo')
    c = list(await repo.execute(q))
    assert k1 in [e.key for e in c]


@pytest.mark.asyncio
async def test_put_named_and_retrieve(
    client: Client
):
    repo = BaseDatastoreRepository(client=client)
    e1 = repo.entity('Test', secrets.token_urlsafe(32))
    e1['foo'] = 'foo'
    k1 = await repo.put(e1)
    assert k1 == e1.key
    e2 = await repo.get_entity_by_key(e1.key)
    assert e1 == e2


@pytest.mark.asyncio
async def test_put_named_and_query(
    client: Client
):
    repo = BaseDatastoreRepository(client=client)
    e1 = repo.entity('Test', secrets.token_urlsafe(32))
    e1['foo'] = 'foo'
    k1 = await repo.put(e1)
    q = repo.query(kind='Test')
    q.add_filter('foo', '=', 'foo')
    c = list(await repo.execute(q))
    assert k1 in [e.key for e in c]


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest
import pytest_asyncio
from headless.core import httpx

import cbra.core as cbra


class AuthenticationRequiredEndpoint(cbra.Endpoint):
    require_authentication: bool = True
    async def head(self): assert False
    async def get(self): assert False
    async def put(self): assert False
    async def post(self): assert False
    async def patch(self): assert False
    async def delete(self): assert False
    async def trace(self): assert False


@pytest_asyncio.fixture # type: ignore
async def client():
    app = cbra.Application()
    AuthenticationRequiredEndpoint.add_to_router(app, path='/')
    async with httpx.Client(base_url='https://cbra', app=app) as client:
        yield client


@pytest.mark.parametrize("method", [
    "HEAD",
    "GET",
    "POST",
    "PUT",
    "PATCH",
    "DELETE",
    "TRACE"
])
@pytest.mark.asyncio
async def test_require_authentication(
    client: httpx.Client,
    method: str
):
    response = await client.request(
        method=method,
        url='/',
    )
    assert response.status_code == 401


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeVar

import pydantic
import fastapi


T = TypeVar('T', bound='ResourceIdentifier')


class ResourceIdentifier(pydantic.BaseModel):

    @classmethod
    def depends(cls: type[T]) -> None | T:
        return fastapi.Depends(cls.frompath)

    @classmethod
    def frompath(
        cls: type[T],
        request: fastapi.Request
    ) -> None | T:
        try:
            return cls.parse_obj(request.path_params)
        except pydantic.ValidationError:
            return None


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import cbra.core as cbra
from .bookpublication import BookPublication


class Book(cbra.ResourceModel):
    id: int | None = cbra.Field(
        default=None,
        read_only=True,
        path_alias='book_id',
        primary_key=True
    )
    title: str
    publications: list[BookPublication]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .authorizationrequest import AuthorizationRequest
from .authorizationlifecycle import AuthorizationLifecycle
from .iauthorizationserverstorage import IAuthorizationServerStorage
from .jarmauthorizeresponse import JARMAuthorizeResponse
from .queryauthorizeresponse import QueryAuthorizeResponse
from .loginresponse import LoginResponse
from .redirecturi import RedirectURI
from .redirectparameters import RedirectParameters
from .responsemodenotsupported import ResponseModeNotSupported
from .responsevalidationfailure import ResponseValidationFailure
from .unsupportedauthorizationresponse import UnsupportedAuthorizationResponse


__all__: list[str] = [
    'AuthorizationLifecycle',
    'AuthorizationRequest',
    'IAuthorizationServerStorage',
    'JARMAuthorizeResponse',
    'LoginResponse',
    'QueryAuthorizeResponse',
    'RedirectURI',
    'RedirectParameters',
    'ResponseModeNotSupported',
    'ResponseValidationFailure',
    'UnsupportedAuthorizationResponse',
]